#ifndef MSP_GEOMETRY_SHAPE_H_
#define MSP_GEOMETRY_SHAPE_H_

#include <vector>
#include <msp/linal/vector.h>
#include "boundingbox.h"
#include "ray.h"
#include "surfacepoint.h"

namespace Msp {
namespace Geometry {

/**
Base class and interface for geometric shapes.  Shapes may be bounded or
unbounded.  They are always considered to be solid, i.e. have a distinct inside
and an outside.
*/
template<typename T, unsigned D>
class Shape
{
protected:
	Shape() { }
public:
	virtual ~Shape() { }

	virtual Shape *clone() const = 0;

	virtual BoundingBox<T, D> get_axis_aligned_bounding_box() const = 0;
	virtual bool contains(const LinAl::Vector<T, D> &) const = 0;
	bool check_intersection(const Ray<T, D> &) const;
	virtual unsigned get_max_ray_intersections() const = 0;
	virtual unsigned get_intersections(const Ray<T, D> &, SurfacePoint<T, D> *, unsigned) const = 0;
	std::vector<SurfacePoint<T, D> > get_intersections(const Ray<T, D> &) const;
};

template<typename T, unsigned D>
inline bool Shape<T, D>::check_intersection(const Ray<T, D> &ray) const
{
	return get_intersections(ray, 0, 1);
}

template<typename T, unsigned D>
inline std::vector<SurfacePoint<T, D> > Shape<T, D>::get_intersections(const Ray<T, D> &ray) const
{
	unsigned max_isect = get_max_ray_intersections();
	std::vector<SurfacePoint<T, D> > points(max_isect);
	unsigned count = get_intersections(ray, &points[0], max_isect);
	points.resize(count);
	return points;
}

} // namespace Geometry
} // namespace Msp

#endif
